@@ -21,6 +21,7 @@
 #include <algorithm>
 #include <assert.h>
 #include <atomic>
+#include <functional>
 #include <iterator>
 #include <mutex>
 #include <sys/types.h>
@@ -77,16 +78,32 @@ namespace REHex
 			
 			std::vector< std::pair<Range, T> > ranges;
 			
+			/* Last iterator returned by get_range(), used to avoid a lookup from
+			 * scratch when a nearby offset is requested again.
+			*/
+			mutable typename std::vector< std::pair<Range, T> >::const_iterator last_get_iter;
+			
 		public:
 			/**
 			 * @brief Construct an empty map.
 			*/
 			ByteRangeMap(const T &default_value = T()):
-				default_value(default_value) {}
+				default_value(default_value),
+				last_get_iter(ranges.end()) {}
 			
 			ByteRangeMap(const ByteRangeMap &src):
 				default_value(src.default_value),
-				ranges(src.ranges) {}
+				ranges(src.ranges),
+				last_get_iter(ranges.end()) {}
+			
+			ByteRangeMap &operator=(const ByteRangeMap &rhs)
+			{
+				default_value = rhs.default_value;
+				ranges = rhs.ranges;
+				last_get_iter = ranges.end();
+				
+				return *this;
+			}
 			
 			bool operator==(const ByteRangeMap<T> &rhs) const
 			{
@@ -106,7 +123,8 @@ namespace REHex
 			*/
 			template<typename I> ByteRangeMap(const I begin, const I end, const T &default_value = T()):
 				default_value(default_value),
-				ranges(begin, end) {}
+				ranges(begin, end),
+				last_get_iter(ranges.end()) {}
 			
 			/**
 			 * @brief Search the map for a range encompassing the given offset.
@@ -141,6 +159,35 @@ namespace REHex
 			*/
 			void clear_range(off_t offset, off_t length);
 			
+			/**
+			 * @brief Get a subset of the ranges defined in the map.
+			 *
+			 * This method builds a ByteRangeMap containing any ranges intersecting the
+			 * given range, clamped to the ends of the range.
+			*/
+			ByteRangeMap<T> get_slice(off_t offset, off_t length) const;
+			
+			/**
+			 * @brief Set all keys defined in another ByteRangeMap.
+			 *
+			 * This method copies ranges from another ByteRangeMap, overwriting any
+			 * already set in this one.
+			*/
+			void set_slice(const ByteRangeMap<T> &slice);
+			
+			/**
+			 * @brief Transform all values defined in the map.
+			 *
+			 * This method transforms ALL values in the map through the given function.
+			 *
+			 * The existing map is modified and a reference to it is returned as a
+			 * convenience to allow for chaining.
+			 *
+			 * WARNING: The transform MUST NOT cause values that were previously equal
+			 * to become not-equal or vice-versa.
+			*/
+			ByteRangeMap<T> &transform(const std::function<T(const T &value)> &func);
+			
 			/**
 			 * @brief Get a reference to the internal std::vector.
 			*/
@@ -152,6 +199,9 @@ namespace REHex
 			const_iterator begin() const { return ranges.begin(); }
 			const_iterator end() const { return ranges.end(); }
 			bool empty() const { return ranges.empty(); }
+			size_t size() const { return ranges.size(); }
+			const std::pair<Range, T> &front() const { assert(!ranges.empty()); return ranges.front(); }
+			const std::pair<Range, T> &back() const { assert(!ranges.empty()); return ranges.back(); }
 			
 			/**
 			 * @brief Adjust for data being inserted into file.
@@ -179,6 +229,11 @@ namespace REHex
 
 template<typename T> typename REHex::ByteRangeMap<T>::const_iterator REHex::ByteRangeMap<T>::get_range(off_t offset) const
 {
+	if(last_get_iter != ranges.end() && last_get_iter->first.offset <= offset && (last_get_iter->first.offset + last_get_iter->first.length) > offset)
+	{
+		return last_get_iter;
+	}
+	
 	/* Starting from the first element after us (or the end of the vector)... */
 	auto i = std::lower_bound(ranges.begin(), ranges.end(), std::make_pair(Range((offset + 1), 0), default_value));
 	
@@ -191,6 +246,7 @@ template<typename T> typename REHex::ByteRangeMap<T>::const_iterator REHex::Byte
 		if(i->first.offset <= offset && (i->first.offset + i->first.length) > offset)
 		{
 			/* ...it does, return it. */
+			last_get_iter = i;
 			return i;
 		}
 	}
@@ -208,12 +264,13 @@ template<typename T> typename REHex::ByteRangeMap<T>::const_iterator REHex::Byte
 		--i;
 	}
 	
-	for(; i != ranges.end() && i->first.offset < (offset + length); ++i)
+	off_t end = offset + length;
+	
+	for(; i != ranges.end() && (i->first.offset < end || end < offset); ++i)
 	{
-		off_t end = offset + length;
 		off_t i_end = i->first.offset + i->first.length;
 		
-		if(i->first.offset < end && offset < i_end)
+		if((i->first.offset < end || end < offset) && offset < i_end)
 		{
 			return i;
 		}
@@ -340,6 +397,8 @@ template<typename T> void REHex::ByteRangeMap<T>::set_range(off_t offset, off_t
 		erase_end = ranges.insert(erase_end, insert_after.front());
 		++erase_end;
 	}
+	
+	last_get_iter = ranges.end();
 }
 
 template<typename T> void REHex::ByteRangeMap<T>::clear_range(off_t offset, off_t length)
@@ -416,6 +475,43 @@ template<typename T> void REHex::ByteRangeMap<T>::clear_range(off_t offset, off_
 		erase_end = ranges.insert(erase_end, insert_after.front());
 		++erase_end;
 	}
+	
+	last_get_iter = ranges.end();
+}
+
+template<typename T> REHex::ByteRangeMap<T> REHex::ByteRangeMap<T>::get_slice(off_t offset, off_t length) const
+{
+	off_t end = offset + length;
+	
+	ByteRangeMap<T> slice;
+	
+	for(auto i = get_range_in(offset, length); i != this->end() && i->first.offset < end; ++i)
+	{
+		off_t slice_off = std::max(i->first.offset, offset);
+		off_t slice_len = std::min(i->first.length, (end - slice_off));
+		
+		slice.set_range(slice_off, slice_len, i->second);
+	}
+	
+	return slice;
+}
+
+template<typename T> void REHex::ByteRangeMap<T>::set_slice(const ByteRangeMap<T> &slice)
+{
+	for(auto i = slice.begin(); i != slice.end(); ++i)
+	{
+		set_range(i->first.offset, i->first.length, i->second);
+	}
+}
+
+template<typename T> REHex::ByteRangeMap<T> &REHex::ByteRangeMap<T>::transform(const std::function<T(const T &value)> &func)
+{
+	for(auto i = ranges.begin(); i != ranges.end(); ++i)
+	{
+		i->second = func(i->second);
+	}
+	
+	return *this;
 }
 
 template<typename T> bool REHex::ByteRangeMap<T>::data_inserted(off_t offset, off_t length)
@@ -506,6 +602,8 @@ template<typename T> bool REHex::ByteRangeMap<T>::data_inserted(off_t offset, of
 		ranges.insert(std::next(ranges.begin(), insert_idx), insert_elem[0]);
 	}
 	
+	last_get_iter = ranges.end();
+	
 	return elements_changed;
 }
 
@@ -591,6 +689,8 @@ template<typename T> bool REHex::ByteRangeMap<T>::data_erased(off_t offset, off_
 		elements_changed = true;
 	}
 	
+	last_get_iter = ranges.end();
+	
 	return elements_changed;
 }
 
