@@ -1,5 +1,5 @@
 /* Reverse Engineer's Hex Editor
- * Copyright (C) 2017-2020 Daniel Collins <solemnwarning@solemnwarning.net>
+ * Copyright (C) 2017-2021 Daniel Collins <solemnwarning@solemnwarning.net>
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published by
@@ -29,8 +29,10 @@
 #include <wx/wx.h>
 
 #include "buffer.hpp"
+#include "ByteRangeSet.hpp"
 #include "document.hpp"
 #include "Events.hpp"
+#include "LRUCache.hpp"
 #include "NestedOffsetLengthMap.hpp"
 #include "Palette.hpp"
 #include "SharedDocumentPointer.hpp"
@@ -132,7 +134,7 @@ namespace REHex {
 					};
 					
 					static void draw_hex_line(DocumentCtrl *doc_ctrl, wxDC &dc, int x, int y, const unsigned char *data, size_t data_len, unsigned int pad_bytes, off_t base_off, bool alternate_row, const std::function<Highlight(off_t)> &highlight_at_off);
-					static void draw_ascii_line(DocumentCtrl *doc_ctrl, wxDC &dc, int x, int y, const unsigned char *data, size_t data_len, unsigned int pad_bytes, off_t base_off, bool alternate_row, const std::function<Highlight(off_t)> &highlight_at_off);
+					static void draw_ascii_line(DocumentCtrl *doc_ctrl, wxDC &dc, int x, int y, const unsigned char *data, size_t data_len, size_t data_extra_pre, size_t data_extra_post, off_t alignment_hint, unsigned int pad_bytes, off_t base_off, bool alternate_row, const std::function<Highlight(off_t)> &highlight_at_off);
 					
 					/**
 					 * @brief Calculate offset of byte at X co-ordinate.
@@ -426,22 +428,114 @@ namespace REHex {
 			
 			void set_cursor_position(off_t position, Document::CursorState cursor_state = Document::CSTATE_GOTO);
 			
+			bool has_prev_cursor_position() const;
+			void goto_prev_cursor_position();
+			
+			bool has_next_cursor_position() const;
+			void goto_next_cursor_position();
+			
 			bool get_insert_mode();
 			void set_insert_mode(bool enabled);
 			
 			void linked_scroll_insert_self_after(DocumentCtrl *p);
 			void linked_scroll_remove_self();
 			
-			void set_selection(off_t off, off_t length);
+			/**
+			 * @brief Set the selection range.
+			 *
+			 * @param begin Data offset at beginning of selection.
+			 * @param end Data offset at end of selection (inclusive).
+			*/
+			bool set_selection_raw(off_t begin, off_t end);
+			
+			/**
+			 * @brief Clear the selection (if any).
+			*/
 			void clear_selection();
-			std::pair<off_t, off_t> get_selection();
+			
+			/**
+			 * @brief Returns true if there is a selection.
+			*/
+			bool has_selection();
+			
+			/**
+			 * @brief Returns the "raw" selection as a begin and end offset.
+			 *
+			 * NOTE: Unlike most "end" pointers, the end offset returned from this
+			 * method is the last byte in the selection, not one past it.
+			*/
+			std::pair<off_t, off_t> get_selection_raw();
+			
+			/**
+			 * @brief Returns the subset of the current selection scoped to a region.
+			 *
+			 * The return value from this method is the offset (file relative) and the
+			 * length of the current selection, scoped to the given region.
+			 *
+			 * If there is no selection, or the selection doesn't include any bytes
+			 * from the given region, the returned length will be <= 0.
+			*/
+			std::pair<off_t, off_t> get_selection_in_region(GenericDataRegion *region);
+			
+			/**
+			 * @brief Returns the set of all bytes currently selected.
+			 *
+			 * NOTE: This method may be expensive to call, as it potentially has to
+			 * iterate through all (data) regions in the file.
+			*/
+			OrderedByteRangeSet get_selection_ranges();
+			
+			/**
+			 * @brief Returns the offset and length of the selection, if linear.
+			 *
+			 * If there is no selection, or the selection isn't linear and contiguous, the length
+			 * will be zero.
+			*/
+			std::pair<off_t, off_t> get_selection_linear();
 			
 			const std::vector<Region*> &get_regions() const;
+			const std::vector<GenericDataRegion*> &get_data_regions() const;
 			void replace_all_regions(std::vector<Region*> &new_regions);
 			bool region_OnChar(wxKeyEvent &event);
 			GenericDataRegion *data_region_by_offset(off_t offset);
 			std::vector<Region*>::iterator region_by_y_offset(int64_t y_offset);
 			
+			/**
+			 * @brief Compare two offsets in the address space defined by the regions.
+			 *
+			 * Returns zero if the two offsets are equal, a negative integer if a is
+			 * less than b and a positive integer if a is greater than b.
+			 *
+			 * Throws an exception of type std::invalid_argument if either of the
+			 * offsets are invalid.
+			*/
+			int region_offset_cmp(off_t a, off_t b);
+			
+			/**
+			 * @brief Increment an offset in the address space defined by the regions.
+			 *
+			 * @param base Base offset to start at.
+			 * @param add Number of bytes to increment base by.
+			 *
+			 * @return New offset, negative if invalid.
+			*/
+			off_t region_offset_add(off_t base, off_t add);
+			
+			/**
+			 * @brief Decrement an offset in the address space defined by the regions.
+			 *
+			 * @param base Base offset to start at.
+			 * @param add Number of bytes to decrement base by.
+			 *
+			 * @return New offset, negative if invalid.
+			*/
+			off_t region_offset_sub(off_t base, off_t sub);
+			
+			/**
+			 * @brief Check if a range of offsets is linear and contiguous.
+			*/
+			bool region_range_linear(off_t begin_offset, off_t end_offset_incl);
+			
 			wxFont &get_font();
 			
 			/**
@@ -452,7 +546,7 @@ namespace REHex {
 			/**
 			 * @brief Set the vertical scroll position, in lines.
 			*/
-			void set_scroll_yoff(int64_t scroll_yoff);
+			void set_scroll_yoff(int64_t scroll_yoff, bool update_linked_scroll_others = true);
 			
 			void OnPaint(wxPaintEvent &event);
 			void OnErase(wxEraseEvent& event);
@@ -528,8 +622,13 @@ namespace REHex {
 			off_t cpos_off{0};
 			bool insert_mode{false};
 			
-			off_t selection_off;
-			off_t selection_length;
+			static const size_t CPOS_HISTORY_LIMIT = 128;
+			
+			std::vector<off_t> cpos_prev;
+			std::vector<off_t> cpos_next;
+			
+			off_t selection_begin;
+			off_t selection_end;
 			
 			bool cursor_visible;
 			wxTimer redraw_cursor_timer;
@@ -544,7 +643,7 @@ namespace REHex {
 			
 			Document::CursorState cursor_state;
 			
-			void _set_cursor_position(off_t position, Document::CursorState cursor_state);
+			void _set_cursor_position(off_t position, Document::CursorState cursor_state, bool preserve_cpos_hist = false);
 			
 			std::vector<GenericDataRegion*>::iterator _data_region_by_offset(off_t offset);
 			
@@ -560,15 +659,68 @@ namespace REHex {
 			void _update_vscroll();
 			void _update_vscroll_pos(bool update_linked_scroll_others = true);
 			
+			/**
+			 * @brief Fuzzy description of the DocumentCtrl scroll position.
+			 *
+			 * This struct describes the DocumentCtrl scroll position, in terms of its
+			 * contents so it can be restored (as close as possible) when the window
+			 * size changes or regions are added/removed/grow/shrink/etc.
+			*/
+			struct FuzzyScrollPosition
+			{
+				bool data_offset_valid;   /**< True if data_offset and data_offset_line are valid. */
+				off_t data_offset;        /**< File offset used as reference point. */
+				int64_t data_offset_line; /**< Visible (on-screen) line where data_offset is. */
+				
+				bool region_idx_valid;    /**< True if region_idx and region_idx_line are valid. */
+				size_t region_idx;        /**< Index of region whose first line is our reference point. */
+				int64_t region_idx_line;  /**< Visible (on-screen) line where region begins (may be negative). */
+				
+				FuzzyScrollPosition():
+					data_offset_valid(false),
+					region_idx_valid(false) {}
+			};
+			
+			/**
+			 * @brief Set scroll_yoff, clamped to valid range.
+			*/
+			void set_scroll_yoff_clamped(int64_t scroll_yoff);
+			
+			/**
+			 * @brief Fetch the current scroll position.
+			*/
+			FuzzyScrollPosition get_scroll_position_fuzzy();
+			
+			/**
+			 * @brief Jump to a fuzzy scroll position.
+			*/
+			void set_scroll_position_fuzzy(const FuzzyScrollPosition &fsp);
+			
+			FuzzyScrollPosition saved_scroll_position;
+			
+			/**
+			 * @brief Save the current scroll position.
+			*/
+			void save_scroll_position();
+			
+			/**
+			 * @brief Restore the last saved scroll position.
+			*/
+			void restore_scroll_position();
+			
 			void linked_scroll_visit_others(const std::function<void(DocumentCtrl*)> &func);
 			
 			static const int PRECOMP_HF_STRING_WIDTH_TO = 512;
 			unsigned int hf_string_width_precomp[PRECOMP_HF_STRING_WIDTH_TO];
 			
+			static const size_t GETTEXTEXTENT_CACHE_SIZE = 4096;
+			LRUCache<std::string, wxSize> hf_gte_cache;
+			
 		public:
 			static std::list<wxString> format_text(const wxString &text, unsigned int cols, unsigned int from_line = 0, unsigned int max_lines = -1);
 			int indent_width(int depth);
 			int get_offset_column_width();
+			int get_virtual_width();
 			bool get_cursor_visible();
 			
 			int hf_char_width();
