@@ -1,5 +1,5 @@
 /* Reverse Engineer's Hex Editor
- * Copyright (C) 2018 Daniel Collins <solemnwarning@solemnwarning.net>
+ * Copyright (C) 2018-2021 Daniel Collins <solemnwarning@solemnwarning.net>
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published by
@@ -21,6 +21,7 @@
 #include <assert.h>
 #include <inttypes.h>
 
+#include "App.hpp"
 #include "decodepanel.hpp"
 #include "Events.hpp"
 
@@ -110,6 +111,9 @@ REHex::DecodePanel::DecodePanel(wxWindow *parent, SharedDocumentPointer &documen
 	pgrid = new wxPropertyGrid(this, wxID_ANY, wxDefaultPosition, wxDefaultSize,
 		wxPG_STATIC_SPLITTER);
 	
+	wxGetApp().Bind(PALETTE_CHANGED, [this](wxCommandEvent &event) { set_pgrid_colours(); });
+	set_pgrid_colours();
+	
 	pgrid->Append(c8 = new wxPropertyCategory("8 bit integer"));
 	pgrid->AppendIn(c8, (s8 = new wxStringProperty("S Dec", "s8", "0000000000000000")));
 	pgrid->AppendIn(c8, (u8 = new wxStringProperty("U Dec", "u8", "0000000000000000")));
@@ -207,6 +211,20 @@ void REHex::DecodePanel::load_state(wxConfig *config)
 	update();
 }
 
+void REHex::DecodePanel::set_pgrid_colours()
+{
+	pgrid->SetCaptionBackgroundColour(    (*active_palette)[Palette::PAL_COMMENT_BG]         );
+	pgrid->SetCaptionTextColour(          (*active_palette)[Palette::PAL_ALTERNATE_TEXT_FG]  );
+	pgrid->SetCellBackgroundColour(       (*active_palette)[Palette::PAL_NORMAL_TEXT_BG]     );
+	pgrid->SetCellDisabledTextColour(     (*active_palette)[Palette::PAL_ALTERNATE_TEXT_FG]  );
+	pgrid->SetCellTextColour(             (*active_palette)[Palette::PAL_NORMAL_TEXT_FG]     );
+	pgrid->SetEmptySpaceColour(           (*active_palette)[Palette::PAL_NORMAL_TEXT_BG]     );
+	pgrid->SetLineColour(                 (*active_palette)[Palette::PAL_COMMENT_BG]         );
+	pgrid->SetMarginColour(               (*active_palette)[Palette::PAL_COMMENT_BG]         );
+	pgrid->SetSelectionBackgroundColour(  (*active_palette)[Palette::PAL_SELECTED_TEXT_BG]   );
+	pgrid->SetSelectionTextColour(        (*active_palette)[Palette::PAL_SELECTED_TEXT_FG]   );
+}
+
 wxSize REHex::DecodePanel::DoGetBestClientSize() const
 {
 	return wxSize(pgrid_best_width, -1);
@@ -424,7 +442,8 @@ void REHex::DecodePanel::OnPropertyGridSelected(wxPropertyGridEvent &event)
 	
 	if(size > 0 && document_ctrl != NULL)
 	{
-		document_ctrl->set_selection(document->get_cursor_position(), size);
+		off_t cursor_position = document->get_cursor_position();
+		document_ctrl->set_selection_raw(cursor_position, (cursor_position + size - 1));
 	}
 }
 
