@@ -1,5 +1,5 @@
 /* Reverse Engineer's Hex Editor
- * Copyright (C) 2017-2020 Daniel Collins <solemnwarning@solemnwarning.net>
+ * Copyright (C) 2017-2021 Daniel Collins <solemnwarning@solemnwarning.net>
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published by
@@ -30,6 +30,7 @@
 #include "buffer.hpp"
 #include "ByteRangeMap.hpp"
 #include "ByteRangeSet.hpp"
+#include "CharacterEncoder.hpp"
 #include "NestedOffsetLengthMap.hpp"
 #include "util.hpp"
 
@@ -151,6 +152,11 @@ namespace REHex {
 			*/
 			bool is_byte_dirty(off_t offset) const;
 			
+			/**
+			 * @brief Check if the BUFFER has any pending changes to be saved.
+			*/
+			bool is_buffer_dirty() const;
+			
 			off_t get_cursor_position() const;
 			CursorState get_cursor_state() const;
 			void set_cursor_position(off_t off, CursorState cursor_state = CSTATE_GOTO);
@@ -238,6 +244,8 @@ namespace REHex {
 			*/
 			bool set_data_type(off_t offset, off_t length, const std::string &type);
 			
+			const CharacterEncoder *get_text_encoder(off_t offset) const;
+			
 			bool set_virt_mapping(off_t real_offset, off_t virt_offset, off_t length);
 			void clear_virt_mapping_r(off_t real_offset, off_t length);
 			void clear_virt_mapping_v(off_t virt_offset, off_t length);
@@ -273,12 +281,24 @@ namespace REHex {
 		#ifndef UNIT_TEST
 		private:
 		#endif
-			struct TrackedChange
+			struct TransOpFunc
 			{
-				const char *desc;
+				const std::function<TransOpFunc()> func;
 				
-				std::function< void() > undo;
-				std::function< void() > redo;
+				TransOpFunc(const std::function<TransOpFunc()> &func);
+				TransOpFunc(const TransOpFunc &src);
+				TransOpFunc(TransOpFunc &&src);
+				
+				TransOpFunc operator()() const;
+			};
+			
+			struct Transaction
+			{
+				const std::string desc;
+				
+				bool complete;
+				
+				std::list<TransOpFunc> ops;
 				
 				off_t       old_cpos_off;
 				CursorState old_cursor_state;
@@ -289,17 +309,29 @@ namespace REHex {
 				ByteRangeMap<off_t> old_real_to_virt_segs;
 				ByteRangeMap<off_t> old_virt_to_real_segs;
 				
-				bool old_dirty;
-				ByteRangeSet old_dirty_bytes;
+				Transaction(const std::string &desc, Document *doc):
+					desc(desc),
+					complete(false),
+					
+					old_cpos_off(doc->get_cursor_position()),
+					old_cursor_state(doc->get_cursor_state()),
+					old_comments(doc->get_comments()),
+					old_highlights(doc->get_highlights()),
+					old_types(doc->get_data_types()),
+					old_real_to_virt_segs(doc->get_real_to_virt_segs()),
+					old_virt_to_real_segs(doc->get_virt_to_real_segs()) {}
 			};
 			
+			void transact_step(const TransOpFunc &op, const std::string &desc);
+			
 			Buffer *buffer;
 			std::string filename;
+			bool write_protect;
 			
-			bool dirty;
-			ByteRangeSet dirty_bytes;
-			
-			void set_dirty(bool dirty);
+			unsigned int current_seq;
+			unsigned int buffer_seq;
+			ByteRangeMap<unsigned int> data_seq;
+			unsigned int saved_seq;
 			
 			NestedOffsetLengthMap<Comment> comments;
 			NestedOffsetLengthMap<int> highlights; /* TODO: Change this to a ByteRangeMap. */
@@ -316,24 +348,33 @@ namespace REHex {
 			enum CursorState cursor_state;
 			
 			static const int UNDO_MAX = 64;
-			std::list<REHex::Document::TrackedChange> undo_stack;
-			std::list<REHex::Document::TrackedChange> redo_stack;
+			std::list<Transaction> undo_stack;
+			std::list<Transaction> redo_stack;
 			
 			void _set_cursor_position(off_t position, enum CursorState cursor_state);
 			
-			void _UNTRACKED_overwrite_data(off_t offset, const unsigned char *data, off_t length);
+			void _UNTRACKED_overwrite_data(off_t offset, const unsigned char *data, off_t length, const ByteRangeMap<unsigned int> &data_seq_slice);
 			
-			void _UNTRACKED_insert_data(off_t offset, const unsigned char *data, off_t length);
+			void _UNTRACKED_insert_data(off_t offset, const unsigned char *data, off_t length, const ByteRangeMap<unsigned int> &data_seq_slice);
 			void _update_mappings_data_inserted(off_t offset, off_t length);
 			
 			void _UNTRACKED_erase_data(off_t offset, off_t length);
 			bool _virt_to_real_segs_data_erased(off_t offset, off_t length);
 			
-			void _tracked_overwrite_data(const char *change_desc, off_t offset, const unsigned char *data, off_t length, off_t new_cursor_pos, CursorState new_cursor_state);
-			void _tracked_insert_data(const char *change_desc, off_t offset, const unsigned char *data, off_t length, off_t new_cursor_pos, CursorState new_cursor_state);
-			void _tracked_erase_data(const char *change_desc, off_t offset, off_t length, off_t new_cursor_pos, CursorState new_cursor_state);
-			void _tracked_replace_data(const char *change_desc, off_t offset, off_t old_data_length, const unsigned char *new_data, off_t new_data_length, off_t new_cursor_pos, CursorState new_cursor_state);
-			void _tracked_change(const char *desc, std::function< void() > do_func, std::function< void() > undo_func);
+			TransOpFunc _op_overwrite_undo(off_t offset, std::shared_ptr< std::vector<unsigned char> > old_data, off_t new_cursor_pos, CursorState new_cursor_state);
+			TransOpFunc _op_overwrite_redo(off_t offset, std::shared_ptr< std::vector<unsigned char> > new_data, off_t new_cursor_pos, CursorState new_cursor_state);
+			
+			TransOpFunc _op_insert_undo(off_t offset, off_t length, off_t new_cursor_pos, CursorState new_cursor_state);
+			TransOpFunc _op_insert_redo(off_t offset, std::shared_ptr< std::vector<unsigned char> > data, off_t new_cursor_pos, CursorState new_cursor_state, const ByteRangeMap<unsigned int> &redo_data_seq_slice);
+			
+			TransOpFunc _op_erase_undo(off_t offset, std::shared_ptr< std::vector<unsigned char> > old_data, off_t new_cursor_pos, CursorState new_cursor_state, const ByteRangeMap<unsigned int> &undo_data_seq_slice);
+			TransOpFunc _op_erase_redo(off_t offset, off_t length, off_t new_cursor_pos, CursorState new_cursor_state);
+			
+			TransOpFunc _op_replace_undo(off_t offset, std::shared_ptr< std::vector<unsigned char> > old_data, off_t new_data_length, off_t new_cursor_pos, CursorState new_cursor_state, const ByteRangeMap<unsigned int> &undo_data_seq_slice);
+			TransOpFunc _op_replace_redo(off_t offset, off_t old_data_length, std::shared_ptr< std::vector<unsigned char> > new_data, off_t new_cursor_pos, CursorState new_cursor_state);
+			
+			void _tracked_change(const char *desc, const std::function< void() > &do_func, const std::function< void() > &undo_func);
+			TransOpFunc _op_tracked_change(const std::function< void() > &func, const std::function< void() > &next_func);
 			
 			json_t *_dump_metadata(bool& has_data);
 			void _save_metadata(const std::string &filename);
@@ -364,6 +405,19 @@ namespace REHex {
 			*/
 			off_t buffer_length() const;
 			
+			/**
+			 * @brief Set write protect flag on the file.
+			 *
+			 * If the write protect flag is set, any attempts to modify the file (buffer) DATA
+			 * will be no-ops. Changes to metadata (comments, etc) are still permitted.
+			*/
+			void set_write_protect(bool write_protect);
+			
+			/**
+			 * @get Get the write protect flag state.
+			*/
+			bool get_write_protect() const;
+			
 			/**
 			 * @brief Overwrite a range of bytes in the file.
 			 *
@@ -386,7 +440,7 @@ namespace REHex {
 			 * @param new_cursor_state  New cursor state. Pass CSTATE_CURRENT to not change the cursor state.
 			 * @param change_desc       Description of change for undo history.
 			*/
-			void insert_data(off_t offset, const unsigned char *data, off_t length,                                      off_t new_cursor_pos = -1, CursorState new_cursor_state = CSTATE_CURRENT, const char *change_desc = "change data");
+			void insert_data(off_t offset, const void *data, off_t length,                                      off_t new_cursor_pos = -1, CursorState new_cursor_state = CSTATE_CURRENT, const char *change_desc = "change data");
 			
 			/**
 			 * @brief Erase a range of bytes in the file.
@@ -410,7 +464,23 @@ namespace REHex {
 			 * @param new_cursor_state  New cursor state. Pass CSTATE_CURRENT to not change the cursor state.
 			 * @param change_desc       Description of change for undo history.
 			*/
-			void replace_data(off_t offset, off_t old_data_length, const unsigned char *new_data, off_t new_data_length, off_t new_cursor_pos = -1, CursorState new_cursor_state = CSTATE_CURRENT, const char *change_desc = "change data");
+			void replace_data(off_t offset, off_t old_data_length, const void *new_data, off_t new_data_length, off_t new_cursor_pos = -1, CursorState new_cursor_state = CSTATE_CURRENT, const char *change_desc = "change data");
+			
+			static const off_t WRITE_TEXT_KEEP_POSITION = -1;  /**< Don't move the cursor after writing. */
+			static const off_t WRITE_TEXT_GOTO_NEXT = -2;      /**< Jump to byte following written data. */
+			
+			static const int WRITE_TEXT_OK = 0;
+			static const int WRITE_TEXT_BAD_OFFSET = 1;
+			static const int WRITE_TEXT_SKIPPED = 2;
+			static const int WRITE_TEXT_TRUNCATED = 4;
+			
+			int overwrite_text(off_t offset, const std::string &utf8_text, off_t new_cursor_pos = WRITE_TEXT_GOTO_NEXT, CursorState new_cursor_state = CSTATE_CURRENT, const char *change_desc = "change data");
+			int insert_text(off_t offset, const std::string &utf8_text, off_t new_cursor_pos = WRITE_TEXT_GOTO_NEXT, CursorState new_cursor_state = CSTATE_CURRENT, const char *change_desc = "change data");
+			int replace_text(off_t offset, off_t old_data_length, const std::string &utf8_text, off_t new_cursor_pos = WRITE_TEXT_GOTO_NEXT, CursorState new_cursor_state = CSTATE_CURRENT, const char *change_desc = "change data");
+			
+			void transact_begin(const std::string &desc);
+			void transact_commit();
+			void transact_rollback();
 	};
 	
 	/**
@@ -462,6 +532,47 @@ namespace REHex {
 			*/
 			void set_comments(const std::list<NestedOffsetLengthMap<REHex::Document::Comment>::const_iterator> &comments, off_t base = 0);
 	};
+	
+	/**
+	 * @brief RAII-style Document transaction wrapper.
+	*/
+	class ScopedTransaction
+	{
+		private:
+			Document *doc;
+			bool committed;
+			
+		public:
+			/**
+			 * @brief Opens a new transaction.
+			*/
+			ScopedTransaction(Document *doc, const std::string &desc):
+				doc(doc),
+				committed(false)
+			{
+				doc->transact_begin(desc);
+			}
+			
+			/**
+			 * @brief Rolls back the transaction if not already committed.
+			*/
+			~ScopedTransaction()
+			{
+				if(!committed)
+				{
+					doc->transact_rollback();
+				}
+			}
+			
+			/**
+			 * @brief Complete the transaction.
+			*/
+			void commit()
+			{
+				doc->transact_commit();
+				committed = true;
+			}
+	};
 }
 
 #endif /* !REHEX_DOCUMENT_HPP */
