@@ -1,5 +1,5 @@
 /* Reverse Engineer's Hex Editor
- * Copyright (C) 2017-2020 Daniel Collins <solemnwarning@solemnwarning.net>
+ * Copyright (C) 2017-2021 Daniel Collins <solemnwarning@solemnwarning.net>
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published by
@@ -24,6 +24,7 @@
 #include <wx/dataobj.h>
 #include <wx/event.h>
 #include <wx/filename.h>
+#include <wx/fontenum.h>
 #include <wx/msgdlg.h>
 #include <wx/aui/auibook.h>
 #include <wx/numdlg.h>
@@ -46,10 +47,12 @@
 #include "../res/icon64.h"
 
 #ifdef __APPLE__
+#include "../res/backward32.h"
 #include "../res/document_new32.h"
 #include "../res/document_open32.h"
 #include "../res/document_save32.h"
 #include "../res/document_save_as32.h"
+#include "../res/forward32.h"
 #endif
 
 enum {
@@ -60,8 +63,10 @@ enum {
 	ID_SEARCH_TEXT,
 	ID_SEARCH_BSEQ,
 	ID_SEARCH_VALUE,
+	ID_COMPARE_FILE,
 	ID_GOTO_OFFSET,
 	ID_OVERWRITE_MODE,
+	ID_WRITE_PROTECT,
 	ID_SAVE_VIEW,
 	ID_INLINE_COMMENTS_HIDDEN,
 	ID_INLINE_COMMENTS_FULL,
@@ -87,6 +92,7 @@ enum {
 
 BEGIN_EVENT_TABLE(REHex::MainWindow, wxFrame)
 	EVT_CLOSE(REHex::MainWindow::OnWindowClose)
+	EVT_CHAR_HOOK(REHex::MainWindow::OnCharHook)
 	
 	EVT_MENU(wxID_NEW,        REHex::MainWindow::OnNew)
 	EVT_MENU(wxID_OPEN,       REHex::MainWindow::OnOpen)
@@ -97,6 +103,9 @@ BEGIN_EVENT_TABLE(REHex::MainWindow, wxFrame)
 	EVT_MENU(ID_CLOSE_OTHERS, REHex::MainWindow::OnCloseOthers)
 	EVT_MENU(wxID_EXIT,       REHex::MainWindow::OnExit)
 	
+	EVT_MENU(wxID_BACKWARD, REHex::MainWindow::OnCursorPrev)
+	EVT_MENU(wxID_FORWARD,  REHex::MainWindow::OnCursorNext)
+	
 	EVT_MENU(wxID_FILE1, REHex::MainWindow::OnRecentOpen)
 	EVT_MENU(wxID_FILE2, REHex::MainWindow::OnRecentOpen)
 	EVT_MENU(wxID_FILE3, REHex::MainWindow::OnRecentOpen)
@@ -115,11 +124,14 @@ BEGIN_EVENT_TABLE(REHex::MainWindow, wxFrame)
 	
 	EVT_MENU(ID_FILL_RANGE, REHex::MainWindow::OnFillRange)
 	EVT_MENU(ID_OVERWRITE_MODE, REHex::MainWindow::OnOverwriteMode)
+	EVT_MENU(ID_WRITE_PROTECT, REHex::MainWindow::OnWriteProtect)
 	
 	EVT_MENU(ID_SEARCH_TEXT, REHex::MainWindow::OnSearchText)
 	EVT_MENU(ID_SEARCH_BSEQ,  REHex::MainWindow::OnSearchBSeq)
 	EVT_MENU(ID_SEARCH_VALUE,  REHex::MainWindow::OnSearchValue)
 	
+	EVT_MENU(ID_COMPARE_FILE,  REHex::MainWindow::OnCompareFile)
+	
 	EVT_MENU(ID_GOTO_OFFSET, REHex::MainWindow::OnGotoOffset)
 	
 	EVT_MENU(wxID_CUT,   REHex::MainWindow::OnCut)
@@ -240,13 +252,19 @@ REHex::MainWindow::MainWindow(const wxSize& size):
 		edit_menu->AppendCheckItem(ID_OVERWRITE_MODE, "Overwrite mode\tIns");
 		#endif
 		
+		edit_menu->AppendCheckItem(ID_WRITE_PROTECT, "Write protect file data");
+		
 		edit_menu->AppendSeparator(); /* ---- */
 		
 		edit_menu->Append(ID_SEARCH_TEXT,  "Search for text...");
 		edit_menu->Append(ID_SEARCH_BSEQ,  "Search for byte sequence...");
 		edit_menu->Append(ID_SEARCH_VALUE, "Search for value...");
 		
-		edit_menu->AppendSeparator();
+		edit_menu->AppendSeparator(); /* ---- */
+		
+		edit_menu->Append(ID_COMPARE_FILE, "Compare whole file...\tCtrl-K");
+		
+		edit_menu->AppendSeparator(); /* ---- */
 		
 		edit_menu->Append(ID_GOTO_OFFSET, "Jump to offset...\tCtrl-G");
 		
@@ -337,6 +355,28 @@ REHex::MainWindow::MainWindow(const wxSize& size):
 		
 		view_menu->AppendSeparator(); /* ---- */
 		
+		wxMenu *font_menu = new wxMenu;
+		view_menu->AppendSubMenu(font_menu, "Select font");
+		
+		wxArrayString font_names = wxFontEnumerator::GetFacenames(wxFONTENCODING_SYSTEM, true);
+		
+		for(size_t i = 0; i < font_names.GetCount(); ++i)
+		{
+			std::string font_name = font_names[i].ToStdString();
+			
+			wxMenuItem *itm = font_menu->AppendRadioItem(wxID_ANY, font_name);
+			if(font_name == wxGetApp().get_font_name())
+			{
+				itm->Check(true);
+			}
+			
+			Bind(wxEVT_MENU, [font_name, itm](wxCommandEvent &event)
+			{
+				wxGetApp().set_font_name(font_name);
+				itm->Check(true);
+			}, itm->GetId(), itm->GetId());
+		}
+		
 		view_menu->Append(ID_FSA_INCREASE, "Increase font size");
 		view_menu->Append(ID_FSA_DECREASE, "Decrease font size");
 		
@@ -409,11 +449,21 @@ REHex::MainWindow::MainWindow(const wxSize& size):
 	toolbar->AddTool(wxID_OPEN,   "Open",    wxBITMAP_PNG_FROM_DATA(document_open32));
 	toolbar->AddTool(wxID_SAVE,   "Save",    wxBITMAP_PNG_FROM_DATA(document_save32));
 	toolbar->AddTool(wxID_SAVEAS, "Save As", wxBITMAP_PNG_FROM_DATA(document_save_as32));
+	
+	toolbar->AddSeparator();
+	
+	toolbar->AddTool(wxID_BACKWARD, "Previous cursor position", wxBITMAP_PNG_FROM_DATA(backward32));
+	toolbar->AddTool(wxID_FORWARD,  "Next cursor position",     wxBITMAP_PNG_FROM_DATA(forward32));
 	#else
 	toolbar->AddTool(wxID_NEW,    "New",     artp.GetBitmap(wxART_NEW,          wxART_TOOLBAR));
 	toolbar->AddTool(wxID_OPEN,   "Open",    artp.GetBitmap(wxART_FILE_OPEN,    wxART_TOOLBAR));
 	toolbar->AddTool(wxID_SAVE,   "Save",    artp.GetBitmap(wxART_FILE_SAVE,    wxART_TOOLBAR));
 	toolbar->AddTool(wxID_SAVEAS, "Save As", artp.GetBitmap(wxART_FILE_SAVE_AS, wxART_TOOLBAR));
+	
+	toolbar->AddSeparator();
+	
+	toolbar->AddTool(wxID_BACKWARD, "Previous cursor position", artp.GetBitmap(wxART_GO_BACK,    wxART_TOOLBAR));
+	toolbar->AddTool(wxID_FORWARD,  "Next cursor position",     artp.GetBitmap(wxART_GO_FORWARD, wxART_TOOLBAR));
 	#endif
 	
 	toolbar->Realize();
@@ -527,6 +577,21 @@ void REHex::MainWindow::OnWindowClose(wxCloseEvent &event)
 	event.Skip();
 }
 
+void REHex::MainWindow::OnCharHook(wxKeyEvent &event)
+{
+	int modifiers = event.GetModifiers();
+	int key = event.GetKeyCode();
+	
+	if(modifiers == (wxMOD_CMD | wxMOD_SHIFT) && key == 'K')
+	{
+		Tab *tab = active_tab();
+		tab->compare_selection();
+	}
+	else{
+		event.Skip();
+	}
+}
+
 void REHex::MainWindow::OnNew(wxCommandEvent &event)
 {
 	new_file();
@@ -679,6 +744,18 @@ void REHex::MainWindow::OnExit(wxCommandEvent &event)
 	Close();
 }
 
+void REHex::MainWindow::OnCursorPrev(wxCommandEvent &event)
+{
+	Tab *tab = active_tab();
+	tab->doc_ctrl->goto_prev_cursor_position();
+}
+
+void REHex::MainWindow::OnCursorNext(wxCommandEvent &event)
+{
+	Tab *tab = active_tab();
+	tab->doc_ctrl->goto_next_cursor_position();
+}
+
 void REHex::MainWindow::OnSearchText(wxCommandEvent &event)
 {
 	wxWindow *cpage = notebook->GetCurrentPage();
@@ -721,6 +798,12 @@ void REHex::MainWindow::OnSearchValue(wxCommandEvent &event)
 	tab->search_dialog_register(sd);
 }
 
+void REHex::MainWindow::OnCompareFile(wxCommandEvent &event)
+{
+	Tab *tab = active_tab();
+	tab->compare_whole_file();
+}
+
 void REHex::MainWindow::OnGotoOffset(wxCommandEvent &event)
 {
 	Tab *tab = active_tab();
@@ -763,17 +846,18 @@ void REHex::MainWindow::OnPaste(wxCommandEvent &event)
 		 * region the chance to handle the paste event.
 		*/
 		
-		off_t selection_off, selection_length;
-		std::tie(selection_off, selection_length) = tab->doc_ctrl->get_selection();
-		
-		if(selection_length > 0)
+		if(tab->doc_ctrl->has_selection())
 		{
-			REHex::DocumentCtrl::GenericDataRegion *selection_region = tab->doc_ctrl->data_region_by_offset(selection_off);
+			off_t selection_first, selection_last;
+			std::tie(selection_first, selection_last) = tab->doc_ctrl->get_selection_raw();
+			
+			REHex::DocumentCtrl::GenericDataRegion *selection_region = tab->doc_ctrl->data_region_by_offset(selection_first);
 			assert(selection_region != NULL);
 			
-			assert(selection_region->d_offset <= selection_off);
+			assert(selection_region->d_offset <= selection_last);
+			assert((selection_region->d_offset + (selection_region->d_length)) > selection_first);
 			
-			if((selection_region->d_offset + selection_region->d_length) >= (selection_off + selection_length))
+			if((selection_region->d_offset + selection_region->d_length) > selection_last)
 			{
 				if(selection_region->OnPaste(tab->doc_ctrl))
 				{
@@ -813,7 +897,10 @@ void REHex::MainWindow::OnPaste(wxCommandEvent &event)
 			wxTheClipboard->GetData(data);
 			
 			try {
-				tab->paste_text(data.GetText().ToStdString());
+				wxString clipboard_text = data.GetText();
+				const wxScopedCharBuffer clipboard_utf8 = clipboard_text.utf8_str();
+				
+				tab->paste_text(std::string(clipboard_utf8.data(), clipboard_utf8.length()));
 			}
 			catch(const std::exception &e)
 			{
@@ -848,7 +935,14 @@ void REHex::MainWindow::OnRedo(wxCommandEvent &event)
 void REHex::MainWindow::OnSelectAll(wxCommandEvent &event)
 {
 	Tab *tab = active_tab();
-	tab->doc_ctrl->set_selection(0, tab->doc->buffer_length());
+	
+	DocumentCtrl::GenericDataRegion *first_region = tab->doc_ctrl->get_data_regions().front();
+	DocumentCtrl::GenericDataRegion *last_region = tab->doc_ctrl->get_data_regions().back();
+	
+	off_t first_off = first_region->d_offset;
+	off_t last_off  = last_region->d_offset + last_region->d_length - (last_region->d_length > 0);
+	
+	tab->doc_ctrl->set_selection_raw(first_off, last_off);
 }
 
 void REHex::MainWindow::OnSelectRange(wxCommandEvent &event)
@@ -873,6 +967,27 @@ void REHex::MainWindow::OnOverwriteMode(wxCommandEvent &event)
 	tab->doc_ctrl->set_insert_mode(!event.IsChecked());
 }
 
+void REHex::MainWindow::OnWriteProtect(wxCommandEvent &event)
+{
+	Document *doc = active_document();
+	
+	if(event.IsChecked() && doc->is_buffer_dirty())
+	{
+		std::string msg
+			= "The content of " + doc->get_title() + " has already been modified.\n"
+			+ "Enable write protect to prevent FURTHER changes?";
+		
+		int res = wxMessageBox(msg, "File data modified", (wxYES_NO | wxICON_EXCLAMATION), this);
+		if(res == wxNO)
+		{
+			edit_menu->Check(ID_WRITE_PROTECT, false);
+			return;
+		}
+	}
+	
+	doc->set_write_protect(event.IsChecked());
+}
+
 void REHex::MainWindow::OnSetBytesPerLine(wxCommandEvent &event)
 {
 	Tab *tab = active_tab();
@@ -1012,6 +1127,9 @@ void REHex::MainWindow::OnPalette(wxCommandEvent &event)
 			abort();
 	}
 	
+	wxCommandEvent pc_event(PALETTE_CHANGED);
+	wxGetApp().ProcessEvent(pc_event);
+	
 	Refresh();
 }
 
@@ -1061,7 +1179,8 @@ void REHex::MainWindow::OnSaveView(wxCommandEvent &event)
 	config->SetPath("/");
 	config->Write("theme", wxString(active_palette->get_name()));
 	config->Write("font-size-adjustment", (long)(wxGetApp().get_font_size_adjustment()));
-
+	config->Write("font-name", wxString(wxGetApp().get_font_name()));
+	
 	// Clean out all previous settings
 	config->DeleteGroup("/default-view/");
 	config->SetPath("/default-view/");
@@ -1114,6 +1233,7 @@ void REHex::MainWindow::OnDocumentChange(wxAuiNotebookEvent& event)
 	Tab *tab = active_tab();
 	
 	edit_menu->Check(ID_OVERWRITE_MODE, !tab->doc_ctrl->get_insert_mode());
+	edit_menu->Check(ID_WRITE_PROTECT, tab->doc->get_write_protect());
 	view_menu->Check(ID_SHOW_OFFSETS, tab->doc_ctrl->get_show_offsets());
 	view_menu->Check(ID_SHOW_ASCII,   tab->doc_ctrl->get_show_ascii());
 	
@@ -1181,6 +1301,7 @@ void REHex::MainWindow::OnDocumentChange(wxAuiNotebookEvent& event)
 	_update_status_mode(tab->doc_ctrl);
 	_update_undo(tab->doc);
 	_update_dirty(tab->doc);
+	_update_cpos_buttons(tab->doc_ctrl);
 	
 	/* Show any search dialogs attached to this tab. */
 	tab->unhide_child_windows();
@@ -1287,6 +1408,11 @@ void REHex::MainWindow::OnCursorUpdate(CursorUpdateEvent &event)
 		_update_status_offset(active_tab);
 	}
 	
+	if(event_src == active_tab->doc || event_src == active_tab->doc_ctrl)
+	{
+		_update_cpos_buttons(active_tab->doc_ctrl);
+	}
+	
 	event.Skip();
 }
 
@@ -1398,6 +1524,25 @@ REHex::Document *REHex::MainWindow::active_document()
 	return active_tab()->doc;
 }
 
+void REHex::MainWindow::switch_tab(DocumentCtrl *doc_ctrl)
+{
+	size_t num_tabs = notebook->GetPageCount();
+	
+	for(size_t i = 0; i < num_tabs; ++i)
+	{
+		wxWindow *page = notebook->GetPage(i);
+		
+		auto tab = dynamic_cast<Tab*>(page);
+		assert(tab != NULL);
+		
+		if(tab->doc_ctrl == doc_ctrl)
+		{
+			notebook->SetSelection(i);
+			break;
+		}
+	}
+}
+
 void REHex::MainWindow::_update_status_offset(Tab *tab)
 {
 	off_t off = tab->doc->get_cursor_position();
@@ -1409,24 +1554,23 @@ void REHex::MainWindow::_update_status_offset(Tab *tab)
 
 void REHex::MainWindow::_update_status_selection(REHex::DocumentCtrl *doc_ctrl)
 {
-	std::pair<off_t,off_t> selection = doc_ctrl->get_selection();
-	
-	off_t selection_off    = selection.first;
-	off_t selection_length = selection.second;
-	
-	if(selection_length > 0)
+	if(doc_ctrl->has_selection())
 	{
-		off_t selection_end = (selection_off + selection_length) - 1;
+		off_t selection_first, selection_last;
+		std::tie(selection_first, selection_last) = doc_ctrl->get_selection_raw();
 		
-		std::string from_text = format_offset(selection_off, doc_ctrl->get_offset_display_base(), selection_end);
-		std::string to_text   = format_offset(selection_end, doc_ctrl->get_offset_display_base(), selection_end);
+		std::string from_text = format_offset(selection_first, doc_ctrl->get_offset_display_base(), selection_last);
+		std::string to_text   = format_offset(selection_last,  doc_ctrl->get_offset_display_base(), selection_last);
+		
+		ByteRangeSet selection = doc_ctrl->get_selection_ranges();
+		off_t selection_bytes = selection.total_bytes();
 		
 		char buf[64];
 		snprintf(buf, sizeof(buf), "Selection: %s - %s (%u bytes)",
 			from_text.c_str(),
 			to_text.c_str(),
 			
-			(unsigned int)(selection_length));
+			(unsigned int)(selection_bytes));
 		
 		SetStatusText(buf, 1);
 	}
@@ -1494,6 +1638,14 @@ void REHex::MainWindow::_update_dirty(REHex::Document *doc)
 	notebook->SetPageBitmap(notebook->GetSelection(), (dirty ? notebook_dirty_bitmap : wxNullBitmap));
 }
 
+void REHex::MainWindow::_update_cpos_buttons(DocumentCtrl *doc_ctrl)
+{
+	wxToolBar *toolbar = GetToolBar();
+	
+	toolbar->EnableTool(wxID_BACKWARD, doc_ctrl->has_prev_cursor_position());
+	toolbar->EnableTool(wxID_FORWARD,  doc_ctrl->has_next_cursor_position());
+}
+
 bool REHex::MainWindow::unsaved_confirm()
 {
 	std::vector<wxString> dirty_files;
