@@ -1,5 +1,5 @@
 /* Reverse Engineer's Hex Editor
- * Copyright (C) 2020 Daniel Collins <solemnwarning@solemnwarning.net>
+ * Copyright (C) 2020-2021 Daniel Collins <solemnwarning@solemnwarning.net>
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published by
@@ -17,6 +17,7 @@
 
 #include "../src/platform.hpp"
 #include <gtest/gtest.h>
+#include <stdint.h>
 #include <stdio.h>
 
 #include "../src/ByteRangeSet.hpp"
@@ -33,7 +34,7 @@ using namespace REHex;
 std::ostream& operator<<(std::ostream& os, const ByteRangeSet::Range& range)
 {
 	char buf[128];
-	snprintf(buf, sizeof(buf), "{ offset = %zd, length = %zd }", range.offset, range.length);
+	snprintf(buf, sizeof(buf), "{ offset = %jd, length = %jd }", (intmax_t)(range.offset), (intmax_t)(range.length));
 	
 	return os << buf;
 }
@@ -1222,3 +1223,131 @@ TEST(ByteRangeSet, IntersectionEmptySet)
 	EXPECT_EQ(ByteRangeSet::intersection( EMPTY_SET,     NON_EMPTY_SET ).get_ranges(), EMPTY_RANGE);
 	EXPECT_EQ(ByteRangeSet::intersection( EMPTY_SET,     EMPTY_SET     ).get_ranges(), EMPTY_RANGE);
 }
+
+TEST(ByteRangeSet, FindFirstIn)
+{
+	const std::vector<ByteRangeSet::Range> RANGES = {
+		ByteRangeSet::Range(10, 10),
+		ByteRangeSet::Range(50, 30),
+	};
+	
+	const ByteRangeSet SET(RANGES.begin(), RANGES.end());
+	
+	EXPECT_EQ(SET.find_first_in( 0,   0), SET.end());
+	EXPECT_EQ(SET.find_first_in( 0,  10), SET.end());
+	EXPECT_EQ(SET.find_first_in( 0,  11), std::next(SET.begin(), 0));
+	EXPECT_EQ(SET.find_first_in( 0,  20), std::next(SET.begin(), 0));
+	EXPECT_EQ(SET.find_first_in( 0,  30), std::next(SET.begin(), 0));
+	EXPECT_EQ(SET.find_first_in( 0, 100), std::next(SET.begin(), 0));
+	EXPECT_EQ(SET.find_first_in(10,   0), SET.end());
+	EXPECT_EQ(SET.find_first_in(10,  10), std::next(SET.begin(), 0));
+	EXPECT_EQ(SET.find_first_in(10,  20), std::next(SET.begin(), 0));
+	EXPECT_EQ(SET.find_first_in(19,   1), std::next(SET.begin(), 0));
+	EXPECT_EQ(SET.find_first_in(20,  30), SET.end());
+	EXPECT_EQ(SET.find_first_in(20,  31), std::next(SET.begin(), 1));
+	EXPECT_EQ(SET.find_first_in(20, 100), std::next(SET.begin(), 1));
+	EXPECT_EQ(SET.find_first_in(79, 100), std::next(SET.begin(), 1));
+	EXPECT_EQ(SET.find_first_in(80, 100), SET.end());
+}
+
+TEST(ByteRangeSet, FindLastIn)
+{
+	const std::vector<ByteRangeSet::Range> RANGES = {
+		ByteRangeSet::Range(10, 10),
+		ByteRangeSet::Range(50, 30),
+	};
+	
+	const ByteRangeSet SET(RANGES.begin(), RANGES.end());
+	
+	EXPECT_EQ(SET.find_last_in( 0,   0), SET.end());
+	EXPECT_EQ(SET.find_last_in( 0,  10), SET.end());
+	EXPECT_EQ(SET.find_last_in( 0,  11), std::next(SET.begin(), 0));
+	EXPECT_EQ(SET.find_last_in( 0,  20), std::next(SET.begin(), 0));
+	EXPECT_EQ(SET.find_last_in( 0,  30), std::next(SET.begin(), 0));
+	EXPECT_EQ(SET.find_last_in( 0, 100), std::next(SET.begin(), 1));
+	EXPECT_EQ(SET.find_last_in(10,   0), SET.end());
+	EXPECT_EQ(SET.find_last_in(10,  10), std::next(SET.begin(), 0));
+	EXPECT_EQ(SET.find_last_in(10,  20), std::next(SET.begin(), 0));
+	EXPECT_EQ(SET.find_last_in(19,   1), std::next(SET.begin(), 0));
+	EXPECT_EQ(SET.find_last_in(19,  31), std::next(SET.begin(), 0));
+	EXPECT_EQ(SET.find_last_in(19,  32), std::next(SET.begin(), 1));
+	EXPECT_EQ(SET.find_last_in(20,  30), SET.end());
+	EXPECT_EQ(SET.find_last_in(20,  31), std::next(SET.begin(), 1));
+	EXPECT_EQ(SET.find_last_in(20, 100), std::next(SET.begin(), 1));
+	EXPECT_EQ(SET.find_last_in(79, 100), std::next(SET.begin(), 1));
+	EXPECT_EQ(SET.find_last_in(80, 100), SET.end());
+}
+
+TEST(OrderedByteRangeSet, EmptySet)
+{
+	OrderedByteRangeSet brs;
+	
+	EXPECT_RANGES();
+}
+
+TEST(OrderedByteRangeSet, AddOneRange)
+{
+	OrderedByteRangeSet brs;
+	
+	brs.set_range(10, 20);
+	
+	EXPECT_RANGES(
+		ByteRangeSet::Range(10, 20),
+	);
+}
+
+TEST(OrderedByteRangeSet, AddExclusiveRanges)
+{
+	OrderedByteRangeSet brs;
+	
+	brs.set_range(10, 20);
+	brs.set_range(80, 40);
+	brs.set_range(40, 30);
+	
+	EXPECT_RANGES(
+		ByteRangeSet::Range(10, 20),
+		ByteRangeSet::Range(80, 40),
+		ByteRangeSet::Range(40, 30),
+	);
+}
+
+TEST(OrderedByteRangeSet, AddSameRange)
+{
+	OrderedByteRangeSet brs;
+	
+	brs.set_range(10, 20);
+	brs.set_range(10, 20);
+	
+	EXPECT_RANGES(
+		ByteRangeSet::Range(10, 20),
+	);
+}
+
+TEST(OrderedByteRangeSet, AddOverlappingRanges)
+{
+	OrderedByteRangeSet brs;
+	
+	brs.set_range(10, 20);
+	brs.set_range(15, 40);
+	brs.set_range(20, 10);
+	
+	brs.set_range(80, 10);
+	brs.set_range(70, 15);
+	brs.set_range(85, 10);
+	
+	EXPECT_RANGES(
+		ByteRangeSet::Range(10, 20),
+		ByteRangeSet::Range(30, 25),
+		ByteRangeSet::Range(80, 10),
+		ByteRangeSet::Range(70, 10),
+		ByteRangeSet::Range(90,  5),
+	);
+	
+	EXPECT_FALSE(brs.isset(9));
+	EXPECT_TRUE(brs.isset(10, 45));
+	EXPECT_FALSE(brs.isset(55));
+	
+	EXPECT_FALSE(brs.isset(69));
+	EXPECT_TRUE(brs.isset(70, 25));
+	EXPECT_FALSE(brs.isset(95));
+}
