@@ -1,5 +1,5 @@
 /* Reverse Engineer's Hex Editor
- * Copyright (C) 2020 Daniel Collins <solemnwarning@solemnwarning.net>
+ * Copyright (C) 2020-2021 Daniel Collins <solemnwarning@solemnwarning.net>
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published by
@@ -90,6 +90,7 @@ class StringPanelTest: public ::testing::Test
 TEST_F(StringPanelTest, EmptyFile)
 {
 	string_panel = new StringPanel(&frame, doc, main_doc_ctrl);
+	string_panel->set_min_string_length(4);
 	string_panel->set_visible(true);
 	
 	EXPECT_EQ(string_panel->get_num_threads(), 0U) << "StringPanel doesn't spawn workers for an empty file";
@@ -108,6 +109,7 @@ TEST_F(StringPanelTest, TextOnlyFile)
 	doc->insert_data(0, DATA.data(), DATA.size());
 	
 	string_panel = new StringPanel(&frame, doc, main_doc_ctrl);
+	string_panel->set_min_string_length(4);
 	string_panel->set_visible(true);
 	
 	EXPECT_NE(string_panel->get_num_threads(), 0U) << "StringPanel spawns workers for non-empty file";
@@ -133,6 +135,7 @@ TEST_F(StringPanelTest, BinaryOnlyFile)
 	doc->insert_data(0, DATA.data(), DATA.size());
 	
 	string_panel = new StringPanel(&frame, doc, main_doc_ctrl);
+	string_panel->set_min_string_length(4);
 	string_panel->set_visible(true);
 	
 	EXPECT_TRUE(string_panel->get_num_threads() > 0U) << "StringPanel spawns workers for non-empty file";
@@ -162,6 +165,7 @@ TEST_F(StringPanelTest, MixedFile)
 	doc->insert_data(0, data.data(), data.size());
 	
 	string_panel = new StringPanel(&frame, doc, main_doc_ctrl);
+	string_panel->set_min_string_length(4);
 	string_panel->set_visible(true);
 	
 	EXPECT_NE(string_panel->get_num_threads(), 0U) << "StringPanel spawns workers for non-empty file";
@@ -194,6 +198,7 @@ TEST_F(StringPanelTest, OverwriteDataTruncatesString)
 	doc->overwrite_data(256, "crazy nutty grass", 17);
 	
 	string_panel = new StringPanel(&frame, doc, main_doc_ctrl);
+	string_panel->set_min_string_length(4);
 	string_panel->set_visible(true);
 	
 	wait_for_idle(1000);
@@ -245,6 +250,7 @@ TEST_F(StringPanelTest, OverwriteDataSplitsString)
 	doc->overwrite_data(256, "broad slope peep", 16);
 	
 	string_panel = new StringPanel(&frame, doc, main_doc_ctrl);
+	string_panel->set_min_string_length(4);
 	string_panel->set_visible(true);
 	
 	wait_for_idle(1000);
@@ -297,6 +303,7 @@ TEST_F(StringPanelTest, OverwriteDataSplitsInvalidatesString)
 	doc->overwrite_data(256, "broad slope peep", 16);
 	
 	string_panel = new StringPanel(&frame, doc, main_doc_ctrl);
+	string_panel->set_min_string_length(4);
 	string_panel->set_visible(true);
 	
 	wait_for_idle(1000);
@@ -347,6 +354,7 @@ TEST_F(StringPanelTest, OverwriteDataCompletesString)
 	doc->overwrite_data(128, "abc", 3);
 	
 	string_panel = new StringPanel(&frame, doc, main_doc_ctrl);
+	string_panel->set_min_string_length(4);
 	string_panel->set_visible(true);
 	
 	wait_for_idle(1000);
@@ -396,6 +404,7 @@ TEST_F(StringPanelTest, InsertData)
 	doc->overwrite_data(512, "knowledge spotty identify", 25);
 	
 	string_panel = new StringPanel(&frame, doc, main_doc_ctrl);
+	string_panel->set_min_string_length(4);
 	string_panel->set_visible(true);
 	
 	wait_for_idle(1000);
@@ -451,6 +460,7 @@ TEST_F(StringPanelTest, InsertDataCompletesString)
 	doc->overwrite_data(128, "abc", 3);
 	
 	string_panel = new StringPanel(&frame, doc, main_doc_ctrl);
+	string_panel->set_min_string_length(4);
 	string_panel->set_visible(true);
 	
 	wait_for_idle(1000);
@@ -501,6 +511,7 @@ TEST_F(StringPanelTest, EraseData)
 	doc->overwrite_data(512, "pumped stick feeble",   19);
 	
 	string_panel = new StringPanel(&frame, doc, main_doc_ctrl);
+	string_panel->set_min_string_length(4);
 	string_panel->set_visible(true);
 	
 	wait_for_idle(1000);
@@ -554,6 +565,7 @@ TEST_F(StringPanelTest, EraseDataInvalidate)
 	doc->overwrite_data(256, "sturdy books scrape", 19);
 	
 	string_panel = new StringPanel(&frame, doc, main_doc_ctrl);
+	string_panel->set_min_string_length(4);
 	string_panel->set_visible(true);
 	
 	wait_for_idle(1000);
@@ -604,6 +616,7 @@ TEST_F(StringPanelTest, EraseDataMerge)
 	doc->overwrite_data(256, "kettle kneel supply", 19);
 	
 	string_panel = new StringPanel(&frame, doc, main_doc_ctrl);
+	string_panel->set_min_string_length(4);
 	string_panel->set_visible(true);
 	
 	wait_for_idle(1000);
@@ -654,6 +667,7 @@ TEST_F(StringPanelTest, EraseDataCompletesString)
 	doc->overwrite_data(132, "d", 1);
 	
 	string_panel = new StringPanel(&frame, doc, main_doc_ctrl);
+	string_panel->set_min_string_length(4);
 	string_panel->set_visible(true);
 	
 	wait_for_idle(1000);
@@ -702,6 +716,7 @@ TEST_F(StringPanelTest, BackToBackModifications)
 	doc->insert_data(0, BIN_DATA.data(), 16 * MiB);
 	
 	string_panel = new StringPanel(&frame, doc, main_doc_ctrl);
+	string_panel->set_min_string_length(4);
 	string_panel->set_visible(true);
 	
 	wait_for_idle(5000);
@@ -731,7 +746,7 @@ TEST_F(StringPanelTest, BackToBackModifications)
 	
 	EXPECT_NE(string_panel->get_num_threads(), 0U) << "StringPanel spawned worker threads";
 	
-	wait_for_idle(5000);
+	wait_for_idle(10000);
 	
 	EXPECT_EQ(string_panel->get_clean_bytes(), (off_t)(49 * MiB + 320 * kiB)) << "StringPanel processed all data in file";
 	EXPECT_EQ(string_panel->get_num_threads(), 0U) << "StringPanel workers exited";
@@ -751,3 +766,110 @@ TEST_F(StringPanelTest, BackToBackModifications)
 		EXPECT_EQ(got_strings, EXPECT_STRINGS) << "StringPanel finds strings in result of combined operations";
 	}
 }
+
+TEST_F(StringPanelTest, UTF8)
+{
+	const unsigned char DATA[] = {
+		/* Padding */
+		/* 0x00 */ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+		
+		/* Short ASCII-only string */
+		/* 0x08 */ 'A', 'B', 'C', 0x00, 0x00, 0x00, 0x00, 0x00,
+		
+		/* ASCII-only string */
+		/* 0x10 */ 'A', 'B', 'C', 'D', 'E', 'F', 0x00, 0x00,
+		
+		/* Short (enough bytes, but not enough code points) UTF-8 string */
+		/* 0x18 */ 0xC2, 0xA3, 0xE2, 0x98, 0xAD, 0xE2, 0x98, 0x83,
+		
+		/* Padding */
+		/* 0x20 */ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+		
+		/* UTF-8 string */
+		/* 0x28 */ 0xC3, 0xA8, 0xC3, 0xB4, 0xC3, 0xBC, 0xC3, 0xA1,
+		
+		/* Padding */
+		/* 0x30 */ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+		
+		/* Mixed ASCII/UTF-8 string */
+		/* 0x38 */ 'A', 'B', 0xC3, 0xB4, 0xC3, 0xBC, 0x00, 0x00,
+		
+		/* "Hello" in UTF-16LE */
+		/* 0x40 */ 'H', 0x00,  'e', 0x00,  'l', 0x00,  'l', 0x00,
+		/* 0x48 */ 'o', 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+		
+	};
+	
+	doc->insert_data(0, DATA, sizeof(DATA));
+	
+	string_panel = new StringPanel(&frame, doc, main_doc_ctrl);
+	string_panel->set_encoding("UTF-8");
+	string_panel->set_min_string_length(4);
+	string_panel->set_visible(true);
+	
+	wait_for_idle(1000);
+	
+	ASSERT_EQ(string_panel->get_clean_bytes(), 0x50U);
+	ASSERT_EQ(string_panel->get_num_threads(), 0U);
+	
+	{
+		ByteRangeSet strings = string_panel->get_strings();
+		std::vector<ByteRangeSet::Range> got_strings(strings.begin(), strings.end());
+		
+		const std::vector<ByteRangeSet::Range> EXPECT_STRINGS = {
+			ByteRangeSet::Range(0x10, 6),
+			ByteRangeSet::Range(0x28, 8),
+			ByteRangeSet::Range(0x38, 6),
+		};
+		
+		EXPECT_EQ(got_strings, EXPECT_STRINGS);
+	}
+}
+
+TEST_F(StringPanelTest, UTF16)
+{
+	const unsigned char DATA[] = {
+		/* Padding */
+		/* 0x00 */ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+		
+		/* Short ASCII-only string */
+		/* 0x08 */ 'A', 'B', 'C', 0x00, 0x00, 0x00, 0x00, 0x00,
+		
+		/* ASCII-only string */
+		/* 0x10 */ 'A', 'B', 'C', 'D', 'E', 'F', 0x00, 0x00,
+		
+		/* Padding */
+		/* 0x18 */ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+		
+		/* Mixed ASCII/UTF-8 string */
+		/* 0x20 */ 'A', 'B', 0xC3, 0xB4, 0xC3, 0xBC, 0x00, 0x00,
+		
+		/* "Hello" in UTF-16LE */
+		/* 0x28 */ 'H', 0x00,  'e', 0x00,  'l', 0x00,  'l', 0x00,
+		/* 0x30 */ 'o', 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+		
+	};
+	
+	doc->insert_data(0, DATA, sizeof(DATA));
+	
+	string_panel = new StringPanel(&frame, doc, main_doc_ctrl);
+	string_panel->set_encoding("UTF-16LE");
+	string_panel->set_min_string_length(4);
+	string_panel->set_visible(true);
+	
+	wait_for_idle(1000);
+	
+	ASSERT_EQ(string_panel->get_clean_bytes(), 0x38);
+	ASSERT_EQ(string_panel->get_num_threads(), 0U);
+	
+	{
+		ByteRangeSet strings = string_panel->get_strings();
+		std::vector<ByteRangeSet::Range> got_strings(strings.begin(), strings.end());
+		
+		const std::vector<ByteRangeSet::Range> EXPECT_STRINGS = {
+			ByteRangeSet::Range(0x28, 10),
+		};
+		
+		EXPECT_EQ(got_strings, EXPECT_STRINGS);
+	}
+}
